/* packet-wimax.c
 * Routines for WiMAX R6/R4 dissection
 * Copyright 2007, Nitn Naveen [HUGHES Systique Corporation] <nitin.naveen@hsc.com>
 *
 * $Id: packet-wimax.c,v 1.1 2007/05/01 11:56:02 nitin Exp $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <epan/packet.h>
#include <epan/prefs.h>

/* XXX - The "plain" WiMAX port (2231) can be overridden in the prefs */
#define UDP_PORT_WIMAX 2231

/* Forward declaration we need below */
void proto_reg_handoff_wimax(void);

/* Initialize the protocol and registered fields */
static int proto_wimax = -1;
static int hf_wimax_version = -1;
static int hf_wimax_flags = -1;
static int hf_wimax_flags_tbit = -1;
static int hf_wimax_flags_rbit = -1;
static int hf_wimax_function_type = -1;
static int hf_wimax_op_id_and_msg_type = -1;
static int hf_wimax_op_id = -1;
static int hf_wimax_msg_type = -1;
static int hf_wimax_length = -1;
static int hf_wimax_msid = -1;
static int hf_wimax_fourth_word_reserved = -1;
static int hf_wimax_transaction_id = -1;
static int hf_wimax_fifth_word_reserved = -1;
static int hf_wimax_tlv = -1;
static int hf_wimax_tlv_tag = -1;
static int hf_wimax_tlv_length = -1;
static int hf_wimax_tlv_value = -1;

/* Global sample preference ("controls" display of numbers) */
static guint global_wimax_udp_port = UDP_PORT_WIMAX;

/* Initialize the subtree pointers */
static gint ett_wimax = -1;
static gint ett_wimax_flags = -1;
static gint ett_wimax_op_id_and_msg_type = -1;
static gint ett_wimax_tlv = -1;
static dissector_handle_t wimax_handle;


static void get_func_type(guint func_type, gchar **func_type_text)
{
    switch(func_type)
    {
        case 1:
        *func_type_text = "QoS ";
        break;
        case 2:
        *func_type_text = "HO Cntrl ";
        break;
        case 3:
        *func_type_text = "DP Cntrl ";
        break;
        case 4:
        *func_type_text = "Cntx Delivery";
        break;
        case 5:
        *func_type_text = "R3 Mobility ";
        break;
        case 6:
        *func_type_text = "Paging ";
        break;
        case 7:
        *func_type_text = "RRM ";
        break;
        case 8:
        *func_type_text = "Auth Relay ";
        break;
        case 9:
        *func_type_text = "MS State ";
        break;
        case 10:
        *func_type_text = "ReAuth ";
        break;
        default:
        *func_type_text = "Unknown function type ";
    }
}
static void get_message_type(guint func_type,gchar **func_type_text,
                             guint message_type, gchar **message_type_text)
{
    switch(func_type)
    {
        case 1:
            switch(message_type)
            {
                case 1:
                *message_type_text = "RR Ack ";
                break;
                case 2:
                *message_type_text = "RR Req ";
                break;
                case 3:
                *message_type_text = "RR Rsp ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 2:
            switch(message_type)
            {
                case 1:
                *message_type_text = "HO Ack ";
                break;
                case 2:
                *message_type_text = "HO Complete ";
                break;
                case 3:
                *message_type_text = "HO Cnf ";
                break;
                case 4:
                *message_type_text = "HO Req ";
                break;
                case 5:
                *message_type_text = "HO Rsp ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 3:
            switch(message_type)
            {
                case 1:
                *message_type_text = "Path Dereg Ack ";
                break;
                case 2:
                *message_type_text = "Path Dereg Req ";
                break;
                case 3:
                *message_type_text = "Path Dereg Rsp ";
                break;
                case 4:
                *message_type_text = "Path Modification ack ";
                break;
                case 5:
                *message_type_text = "Path Modification Req ";
                break;
                case 6:
                *message_type_text = "Path Modification Rsp ";
                break;
                case 7:
                *message_type_text = "Path Prereg Ack ";
                break;
                case 8:
                *message_type_text = "Path Prereg Req ";
                break;
                case 9:
                *message_type_text = "Path Prereg Rsp ";
                break;
                case 10:
                *message_type_text = "Path Reg Ack ";
                break;
                case 11:
                *message_type_text = "Path Reg Req ";
                break;
                case 12:
                *message_type_text = "Path Reg Rsp ";
                break;
                case 13:
                *message_type_text = "MS Attachment Req ";
                break;
                case 14:
                *message_type_text = "MS Attachment Rsp ";
                break;
                case 15:
                *message_type_text = "MS Attachment Ack ";
                break;
                case 16:
                *message_type_text = "Key Change Directive ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 4:
            switch(message_type)
            {
                case 1:
                *message_type_text = "Context Rpt ";
                break;
                case 2:
                *message_type_text = "Context Req ";
                break;
                case 3:
                *message_type_text = "Context Ack ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 5:
            switch(message_type)
            {
                case 1:
                *message_type_text = "Anchor DPF HO Req ";
                break;
                case 2:
                *message_type_text = "Anchor DPF HO Trigger ";
                break;
                case 3:
                *message_type_text = "Anchor DPF HO Rsp ";
                break;
                case 4:
                *message_type_text = "Anchor DPF Relocate Req ";
                break;
                case 5:
                *message_type_text = "FA Register Req ";
                break;
                case 6:
                *message_type_text = "FA Register Rsp ";
                break;
                case 7:
                *message_type_text = "Anchor DPF Relocate Rsp ";
                break;
                case 8:
                *message_type_text = "FA Revoke Req ";
                break;
                case 9:
                *message_type_text = "FA Revoke Rsp ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 6:
            switch(message_type)
            {
                case 1:
                *message_type_text = "Initiate Paging Req ";
                break;
                case 2:
                *message_type_text = "Initiate Paging Rsp ";
                break;
                case 3:
                *message_type_text = "LU Cnf ";
                break;
                case 4:
                *message_type_text = "LU Req ";
                break;
                case 5:
                *message_type_text = "LU Rsp ";
                break;
                case 6:
                *message_type_text = "Paging Announce ";
                break;
                case 7:
                *message_type_text = "CMAC Key Count Req ";
                break;
                case 8:
                *message_type_text = "CMAC Key Count Rsp ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 7:
            switch(message_type)
            {
                case 1:
                *message_type_text = "R6 PHY Parameters Req ";
                break;
                case 2:
                *message_type_text = "R6 PHY Parameters Rpt ";
                break;
                case 3:
                *message_type_text = "R4/R6 Spare Capacity Req ";
                break;
                case 4:
                *message_type_text = "R4/R6 Spare Capacity Rpt ";
                break;
                case 5:
                *message_type_text = "R6 Neighbor BS Resource Status Update ";
                break;
                case 6:
                *message_type_text = "R4/R6 Radio Config_update Req ";
                break;
                case 7:
                *message_type_text = "R4/R6 Radio Config_update Rpt ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 8:
            switch(message_type)
            {
                case 1:
                *message_type_text = "AR Authenticated EAP Start ";
                break;
                case 2:
                *message_type_text = "AR Authenticated EAP Transfer ";
                break;
                case 3:
                *message_type_text = "AR EAP Start ";
                break;
                case 4:
                *message_type_text = "AR EAP Transfer ";
                break;
                case 5:
                *message_type_text = "AR EAP Complete ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 9:
            switch(message_type)
            {
                case 1:
                *message_type_text = "IM Entry State Change Req ";
                break;
                case 2:
                *message_type_text = "IM Entry State Change Rsp ";
                break;
                case 3:
                *message_type_text = "IM Exit State Change Req ";
                break;
                case 4:
                *message_type_text = "IM Exit State Change Rsp ";
                break;
                case 5:
                *message_type_text = "NW ReEntry State Change Directive ";
                break;
                case 6:
                *message_type_text = "MS PreAttachment Req ";
                break;
                case 7:
                *message_type_text = "MS PreAttachment Rsp ";
                break;
                case 8:
                *message_type_text = "MS PreAttachment Ack ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        case 10:
            switch(message_type)
            {
                case 1:
                *message_type_text = "AR EAP Start ";
                break;
                case 2:
                *message_type_text = "Key Change Directive ";
                break;
                case 3:
                *message_type_text = "Key Change Cnf ";
                break;
                case 4:
                *message_type_text = "Relocation Cnf ";
                break;
                case 5:
                *message_type_text = "Relocation Confirm Ack ";
                break;
                case 6:
                *message_type_text = "Relocation Notify ";
                break;
                case 7:
                *message_type_text = "Relocation Notify Ack ";
                break;
                case 8:
                *message_type_text = "Relocation Req ";
                break;
                case 9:
                *message_type_text = "Relocation Rsp ";
                break;
                default:
                *message_type_text = " Unknown Message Type ";
            }
        break;
        default:
        *func_type_text = "Unknown function type ";     
    }

}


static void get_tag_name(guint tag, gchar **tag_name)
{
    switch(tag)
    {
        case /* Accept/Reject Indicator*/ 1:
        *tag_name = "Accept/Reject Indicator ";
        break;
        case /* Acconting Extension*/ 2:
        *tag_name = "Acconting Extension ";
        break;
        case /* Action Code*/ 3:
        *tag_name = "Action Code ";
        break;
        case /* Action Time*/ 4:
        *tag_name = "Action Time ";
        break;
        case /* AK*/ 5:
        *tag_name = "AK ";
        break;
        case /* ak_context */ 6:
        *tag_name = "AK Context ";
        break;
        case /* ak_id */ 7:
        *tag_name = "AK ID ";
        break;
        case /* ak_lifetime  */8:
        *tag_name = "AK Lifetime ";
        break;
        case /* ak_sn  */9:
        *tag_name = "AK SN ";
        break;
        case /* anchor_asn_gw_id  */10:
        *tag_name = "Anchor ASN GW ID ";
        break;
         case /* Anchor MM Context  */11:
        *tag_name = "Anchor MM Context ";
        break;
         case /* Anchor PCID  */12:
        *tag_name = "Anchor PCID ";
        break;
         case /* Anchor PC Relocation Destination  */13:
        *tag_name = "Anchor PC Relocation Destination  ";
        break;
         case /* Anchor PC Relocation Request Response  */14:
        *tag_name = "Anchor PC Relocation Request Response ";
        break;
        case /* associate_phsi  */15:
        *tag_name = "Associate PHSI ";
        break;
        case /* anchor_authenticator_id  */16:
        *tag_name = "Anchor Authenticator ID ";
        break;
        case /* Authentication Complete  */17:
        *tag_name = "Authentication Complete ";
        break;
        case /* Authentication Result  */18:
        *tag_name = "Authentication Result ";
        break;
        case /* Authenticator Identifier  */19:
        *tag_name = "Authenticator Identifier ";
        break;
        case /* Auth-IND */20:
        *tag_name = "Auth-IND ";
        break;
        case /* authorization_policy  */21:
        *tag_name = "Authorization Policy ";
        break;
        case /* Available Radio Resource DL  */22:
        *tag_name = "Available Radio Resource DL ";
        break;
        case /* Available Radio Resource UL */23:
        *tag_name = "Available Radio Resource UL ";
        break;
        case /* BE Data Delivery Service  */24:
        *tag_name = "BE Data Delivery Service ";
        break;
        case /* bs_id  */25:
        *tag_name = "BS ID ";
        break;
        case /* bs_info  */26:
        *tag_name = "BS Info ";
        break;
        case /* BS-originated EAP-Start Flag  */27:
        *tag_name = "BS-originated EAP-Start Flag ";
        break;
        case /* Care-Of Address */28:
        *tag_name = "Care-Of Address ";
        break;
        case /* CID  */29:
        *tag_name = "CID ";
        break;
        case /* classifier  */30:
        *tag_name = "Classifier ";
        break;
        case /* classifier_action  */31:
        *tag_name = "Classifier Action ";
        break;
        case /* classifier_rule_priority  */32:
        *tag_name = "Classifier Rule Priority ";
        break;
        case /* classifier_type  */33:
        *tag_name = "Classifier Type ";
        break;
        case /* cmac_key_count  */34:
        *tag_name = "CMAC Key Count ";
        break;
        case /* Combined Resources Required  */35:
        *tag_name = "Combined Resources Required ";
        break;
        case /*Context Purpose Indicator  */36:
        *tag_name = "Context Purpose Indicator ";
        break;
        case /* Correlation ID */37:
        *tag_name = "Correlation ID ";
        break;
        case /* Cryptographic Suite  */38:
        *tag_name = "Cryptographic Suite ";
        break;
        case /* cs_type  */39:
        *tag_name = "CS Type ";
        break;
        case /* data_integrity  */40:
        *tag_name = "Data Integrity ";
        break;
        case /* Data Integrity Info  */41:
        *tag_name = "Data Integrity Info ";
        break;
        case /* data_path_encapsulation_type  */42:
        *tag_name = "Datapath Encap Type ";
        break;
        case /* Data Path Establishment Option  */43:
        *tag_name = "Data Path Establishment Option ";
        break;
        case /* data_path_id  */44:
        *tag_name = "Datapath ID ";
        break;
        case /* data_path_info  */45:
        *tag_name = "Datapath Info ";
        break;
        case /* Data Path Integrity Mechanism  */46:
        *tag_name = "Data Path Integrity Mechanism ";
        break;
        case /* data_path_type  */47:
        *tag_name = "Datapath Type ";
        break;
        case /* DCD/UCD Configuration Change Count */48:
        *tag_name = "DCD/UCD Configuration Change Count ";
        break;
        case /* DCD Setting  */49:
        *tag_name = "DCD Setting ";
        break;
        case /* Device Authentication Indicator  */50:
        *tag_name = "Device Authentication Indicator ";
        break;
        case /* DHCP Key  */51:
        *tag_name = "DHCP Key ";
        break;
        case /* DHCP Key ID  */52:
        *tag_name = "DHCP Key ID ";
        break;
        case /* DHCP Key Lifetime  */53:
        *tag_name = "DHCP Key Lifetime";
        break;
        case /* DHCP Proxy Info */54:
        *tag_name = "DHCP Proxy Info ";
        break;
        case /* DHCP Relay Address  */55:
        *tag_name = "DHCP Relay Address ";
        break;
        case /* DHCP Relay Info  */56:
        *tag_name = "DHCP Relay Info ";
        break;
        case /* DHCP Server Address  */57:
        *tag_name = "DHCP Server Address ";
        break;
        case /* DHCP Server List  */58:
        *tag_name = "DHCP Server List ";
        break;
        case /* direction  */59:
        *tag_name = "Direction ";
        break;
        case /* DL PHY Quality Info  */60:
        *tag_name = "DL PHY Quality Info ";
        break;
        case /* DL PHY Service Level  */61:
        *tag_name = "DL PHY Service Level ";
        break;
        case /* eap_payload  */62:
        *tag_name = "EAP Payload ";
        break;
        case /* eik  */63:
        *tag_name = "EIK ";
        break;
        case /* ERT-VR Data Delivery Service */64:
        *tag_name = "ERT-VR Data Delivery Service ";
        break;
        case /* Exit IDLE Mode Operation Indication  */65:
        *tag_name = "Exit IDLE Mode Operation Indication ";
        break;
        case /* FA-HA Key  */66:
        *tag_name = "FA-HA Key ";
        break;
        case /* FA-HA Key Lifetime  */67:
        *tag_name = "FA-HA Key Lifetime ";
        break;
        case /* FA-HA Key SPI  */68:
        *tag_name = "FA-HA Key SPI ";
        break;
        case /* Failure Indication  */69:
        *tag_name = "Failure Indication ";
        break;
        case /* FA IP Address  */70:
        *tag_name = "FA IP Address ";
        break;
        case /* FA Relocation Indication  */71:
        *tag_name = "FA Relocation Indication ";
        break;
        case /* Full DCD Setting  */72:
        *tag_name = "Full DCD Setting ";
        break;
        case /* Full UCD Setting  */73:
        *tag_name = "Full UCD Setting ";
        break;
        case /* Global Service Class Change  */74:
        *tag_name = "Global Service Class Change ";
        break;
        case /* HA IP Address  */75:
        *tag_name = "HA IP Address";
        break;
        case /* HO Confirm Type  */76:
        *tag_name = "HO Confirm Type ";
        break;
        case /* Home Address  */77:
        *tag_name = "Home Address ";
        break;
        case /* HO Process Optimization  */78:
        *tag_name = "HO Process Optimization ";
        break;
        case /* HO Type */79:
        *tag_name = "HO Type ";
        break;
        case /* IDLE Mode Info  */80:
        *tag_name = "IDLE Mode Info ";
        break;
        case /* IDLE Mode Retain Info  */81:
        *tag_name = "IDLE Mode Retain Info ";
        break;
        case /* ip_dest_addr_and_mask  */82:
        *tag_name = "IP Dest Addr and Mask ";
        break;
        case /* IP Remained Time  */83:
        *tag_name = "IP Remained Time ";
        break;
        case /* ip_src_addr_and_mask  */84:
        *tag_name = "IP Src Addr and Mask ";
        break;
        case /* IP TOS/DSCP Range and Mask  */85:
        *tag_name = "IP TOS/DSCP Range and Mask ";
        break;
        case /* key_change_indi  */86:
        *tag_name = "Key Change Indi ";
        break;
        case /* L-BSID  */87:
        *tag_name = "L-BSID ";
        break;
        case /* Location Update Status  */88:
        *tag_name = "Location Update Status ";
        break;
        case /* Location Update Success/Failure Indication  */89:
        *tag_name = "Location Update Success/Failure Indication ";
        break;
        case /* LU Result Indicator  */90:
        *tag_name = "LU Result Indicator ";
        break;
        case /* Maximum Latency  */91:
        *tag_name = "Maximum Latency ";
        break;
        case /* Maximum Sustained Traffic Rate  */92:
        *tag_name = "Maximum Sustained Traffic Rate ";
        break;
        case /* Maximum Traffic Burst  */93:
        *tag_name = "Maximum Traffic Burst ";
        break;
        case /* Media Flow Type  */94:
        *tag_name = "Media Flow Type ";
        break;
        case /* Minimum Reserved Traffic Rate  */95:
        *tag_name = "Minimum Reserved Traffic Rate ";
        break;
        case /* MIP4 Info  */96:
        *tag_name = "MIP4 Info ";
        break;
        case /* MIP4 Security Info  */97:
        *tag_name = "MIP4 Security Info ";
        break;
        case /* MN-FA Key  */98:
        *tag_name = "MN-FA Key ";
        break;
        case /* MN-FA SPI  */99:
        *tag_name = "MN-FA SPI ";
        break;
        case /* MS Authorization Context  */100:
        *tag_name = "MS Authorization Context ";
        break;
        case /* MS FA Context  */101:
        *tag_name = "MS FA Context ";
        break;
        case /* ms_id  */102:
        *tag_name = "MS ID ";
        break;
        case /* ms_info  */103:
        *tag_name = "MS Info ";
        break;
        case /* MS Mobility Mode  */104:
        *tag_name = "MS Mobility Mode ";
        break;
        case /* ms_nai  */105:
        *tag_name = "MS NAI ";
        break;
        case /* MS Networking Context  */106:
        *tag_name = "MS Networking Context ";
        break;
        case /* MS Security Context  */107:
        *tag_name = "MS Security Context ";
        break;
        case /* MS Security History  */108:
        *tag_name = "MS Security History ";
        break;
        case /* Network Exit Indicator  */109:
        *tag_name = "Network Exit Indicator ";
        break;
        case /* Newer TEK Parameters  */110:
        *tag_name = "Newer TEK Parameters ";
        break;
        case /* NRT-VR Data Delivery Service  */111:
        *tag_name = "NRT-VR Data Delivery Service ";
        break;
        case /* Older TEK Parameters  */112:
        *tag_name = "Older TEK Parameters ";
        break;
        case /* Old Anchor PCID  */113:
        *tag_name = "Old Anchor PCID ";
        break;
        case /* packet_class  */114:
        *tag_name = "Packet Class ";
        break;
        case /* Paging Announce Timer  */115:
        *tag_name = "Paging Announce Timer ";
        break;
        case /* Paging Cause  */116:
        *tag_name = "Paging Cause ";
        break;
        case /* Paging Controller Identifier  */117:
        *tag_name = "Paging Controller Identifier ";
        break;
        case /* Paging Cycle  */118:
        *tag_name = "Paging Cycle ";
        break;
        case /* Paging Information  */119:
        *tag_name = "Paging Information ";
        break;
        case /* Paging Offset  */120:
        *tag_name = "Paging Offset ";
        break;
        case /* Paging Start/Stop  */121:
        *tag_name = "Paging Start/Stop ";
        break;
        case /* PC Relocation Indication  */122:
        *tag_name = "PC Relocation Indication ";
        break;
        case /* PGID - Paging Group ID  */123:
        *tag_name = "PGID - Paging Group ID ";
        break;
        case /* PHSF  */124:
        *tag_name = "PHSF ";
        break;
        case /* PHSI  */125:
        *tag_name = "PHSI ";
        break;
        case /* PHSM  */126:
        *tag_name = "PHSM ";
        break;
        case /* PHS Rule  */127:
        *tag_name = "PHS Rule ";
        break;
        case /* PHS Rule Action  */128:
        *tag_name = "PHS Rule Action ";
        break;
        case /* PHSS  */129:
        *tag_name = "PHSS ";
        break;
        case /* PHSV  */130:
        *tag_name = "PHSV ";
        break;
        case /* PKM Context  */131:
        *tag_name = "PKM Context ";
        break;
        case /* PMIP4 Client Location  */132:
        *tag_name = "PMIP4 Client Location ";
        break;
        case /* PMK SN  */133:
        *tag_name = "PMK SN ";
        break;
        case /* PKM2  */134:
        *tag_name = "PKM2 ";
        break;
        case /* PMK2 SN  */135:
        *tag_name = "PMK2 SN ";
        break;
        case /* PN Counter  */136:
        *tag_name = "PN Counter ";
        break;
        case /* Preamble Index/Sub-channel Index  */137:
        *tag_name = "Preamble Index/Sub-channel Index ";
        break;
        case /* protocol  */138:
        *tag_name = "Protocol ";
        break;
        case /* proto_dest_port_rng  */139:
        *tag_name = "Proto Dest Port Range ";
        break;
        case /* proto_src_range  */140:
        *tag_name = "Proto Prc Port Range ";
        break;
        case /* QoS Parameters  */141:
        *tag_name = "QoS Parameters ";
        break;
        case /* Radio Resource Fluctuation  */142:
        *tag_name = "Radio Resource Fluctuation ";
        break;
        case /* Reduced Resources Code  */143:
        *tag_name = "Reduced Resources Code ";
        break;
        case /* reg_context  */144:
        *tag_name = "Reg Context ";
        break;
        case /* registration_type  */145:
        *tag_name = "Registration Type ";
        break;
        case /* Relative Delay  */146:
        *tag_name = "Relative Delay ";
        break;
        case /* Relocation Destination ID  */147:
        *tag_name = "Relocation Destination ID ";
        break;
        case /* Relocation Response  */148:
        *tag_name = "Relocation Response ";
        break;
        case /* Relocation Success Indication  */149:
        *tag_name = "Relocation Success Indication ";
        break;
        case /* Request/Transmission Policy  */150:
        *tag_name = "Request/Transmission Policy ";
        break;
        case /* reserved_action  */151:
        *tag_name = "Reserved Action ";
        break;
        case /* reservation_result  */152:
        *tag_name = "Reservation Result ";
        break;
        case /* Response Code */153:
        *tag_name = "Response Code ";
        break;
        case /* Result Code  */154:
        *tag_name = "Result Code ";
        break;
        case /* context_id        */155:
        *tag_name = "Context ID ";
        break;
        case /* Round Trip Delay        */156:
        *tag_name = "Round Trip Delay ";
        break;
        case /* RRM Absolute Threshold Value        */157:
        *tag_name = "RRM Absolute Threshold Value ";
        break;
        case /* RRM Averaging Time    */158:
        *tag_name = "RRM Averaging Time  ";
        break;
        case /* RRM BS Info        */159:
        *tag_name = "RRM BS Info ";
        break;
        case /* RRM BS-MS PHY Quality Info        */160:
        *tag_name = "RRM BS-MS PHY Quality Info ";
        break;
        case /* RRM Relative Threshold RT        */161:
        *tag_name = "RRM Relative Threshold RT ";
        break;
        case /* RRM Reporting Characteristics        */162:
        *tag_name = "RRM Reporting Characteristics ";
        break;
        case /* RRM Reporting Period        */163:
        *tag_name = "RRM Reporting Period ";
        break;
        case /* RRM Spare Capacity Report Type        */164:
        *tag_name = "RRM Spare Capacity Report Type ";
        break;
        case /* RT-VR Data Delivery Service        */165:
        *tag_name = "RT-VR Data Delivery Service ";
        break;
        case /* RxPN Counter       */166:
        *tag_name = "RxPN Counter ";
        break;
        case /* R3 Operation Status        */167:
        *tag_name = "R3 Operation Status ";
        break;
        case /* release_reason         */168:
        *tag_name = "Release Reason ";
        break;
        case /* SAID         */169:
        *tag_name = "SAID ";
        break;
        case /* SA Descriptor         */170:
        *tag_name = "SA Descriptor ";
        break;
        case /* SA Index         */171:
        *tag_name = "SA Index ";
        break;
        case /* SA Service Type         */172:
        *tag_name = "SA Service Type ";
        break;
        case /* SA Type         */173:
        *tag_name = "SA Type ";
        break;
        case /* SBC Context         */174:
        *tag_name = "SBC Context ";
        break;
        case /* SDU BSN Map         */175:
        *tag_name = "SDU BSN Map ";
        break;
        case /* SDU Info         */176:
        *tag_name = "SDU Info ";
        break;
        case /* SDU Size         */177:
        *tag_name = "SDU Size ";
        break;
        case /* SDU SN         */178:
        *tag_name = "SDU SN ";
        break;
        case /* Service Class Name         */179:
        *tag_name = "Service Class Name ";
        break;
        case /* Service Level Prediction         */180:
        *tag_name = "Service Level Prediction ";
        break;
        case /* Service Authorization Code         */181:
        *tag_name = "Service Authorization Code ";
        break;
        case /* Serving/Target Indicator         */182:
        *tag_name = "Serving/Target Indicator ";
        break;
        case /* SF Classification         */183:
        *tag_name = "SF Classification ";
        break;
        case /* sf_id                    */184:
        *tag_name = "SF ID ";
        break;
        case /* sf_info  */185:
        *tag_name = "SF Info ";
        break;
        case /* Spare Capacity Indicator  */186:
        *tag_name = "Spare Capacity Indicator ";
        break;
        case /* TEK  */187:
        *tag_name = "TEK ";
        break;
        case /* TEK Lifetime  */188:
        *tag_name = "TEK Lifetime ";
        break;
        case /* TEK SN  */189:
        *tag_name = "TEK SN ";
        break;
        case /* Tolerated Jitter  */190:
        *tag_name = "Tolerated Jitter ";
        break;
        case /* Total Slots DL  */191:
        *tag_name = "Total Slots DL ";
        break;
        case /* Total Slots UL  */192:
        *tag_name = "Total Slots UL ";
        break;
        case /* Traffic Priority/QoS Priority  */193:
        *tag_name = "Traffic Priority/QoS Priority ";
        break;
        case /* Tunnel Endpoint  */194:
        *tag_name = "Tunnel Endpoint ";
        break;
        case /* UCD Setting  */195:
        *tag_name = "UCD Setting ";
        break;
        case /* UGS Data Delivery Service  */196:
        *tag_name = "UGS Data Delivery Service ";
        break;
        case /* UL PHY Quality Info  */197:
        *tag_name = "UL PHY Quality Info ";
        break;
        case /* UL PHY Service Level  */198:
        *tag_name = "UL PHY Service Level ";
        break;
        case /* Unsolicited Grant Interval  */199:
        *tag_name = "Unsolicited Grant Interval ";
        break;
        case /* Unsolicited Polling Interval  */200:
        *tag_name = "Unsolicited Polling Interval ";
        break;
        case /* VAAA IP Address  */201:
        *tag_name = "VAAA IP Address ";
        break;
        case /* VAAA Realm  */202:
        *tag_name = "VAAA Realm ";
        break;
        case /* dst_mac  */221:
        *tag_name = "DST MAC ";
        break;
        case /* src_mac  */222:
        *tag_name = "SRC MAC ";
        break;
        case /* user_prio  */223:
        *tag_name = "User Prio ";
        break;
        case /* l2_type  */224:
        *tag_name = "L2 Type ";
        break;
        default:
        *tag_name = "";
    }
    return;
}

static int
get_tag_type(guint tag)
{
    switch (tag)
    {
        case 103:
        case 26:
        case 114:
        case 30:
        case 45:
        case 6:
        case 185:
            return 1;
        default:
            ;
    }
    return 0;
}

static int
dissect_wimax_tlv(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree,
                  guint offset, guint remaining_length)
{
    proto_tree *tp = NULL;
    proto_tree *tag_disect_tree = NULL;
    guint tag = 0;
    guint tag_length = 0;
    gchar *tag_name = NULL;
    gchar *tag_data = NULL;
    gchar *tag_data_format = NULL;
    guint tag_data_format_index = 0;
    guint cursor = 0;

    if (0 == remaining_length)
    {
        return 0;
    }

    tag = tvb_get_ntohs(tvb, offset);
    tag_length = tvb_get_ntohs(tvb, (offset + 2));
    if (4 > tag_length)
    {
        /* Error in TLV */
        if (check_col(pinfo->cinfo, COL_INFO))
        {
            get_tag_name(tag, &tag_name);
            col_append_fstr(pinfo->cinfo, COL_INFO,
                            " <Packet is Malformed: Check Tag %s>", tag_name);
            return 1;
        }
        return 1;
    }

    if (tag_length > remaining_length)
    {
        /* Error in TLV */
        if (check_col(pinfo->cinfo, COL_INFO))
        {
            get_tag_name(tag, &tag_name);
            col_append_fstr(pinfo->cinfo, COL_INFO,
                            " <Packet is Malformed: Check Tag %s>", tag_name);
            return 1;
        }
        return 1;
    }
    else
    {
        /* Print the tag */
       get_tag_name(tag, &tag_name);
        tp = proto_tree_add_protocol_format(tree, hf_wimax_tlv, tvb, offset,
                                            tag_length, 
                                            "Tag: %s[%u], Length: %u",
                                            tag_name,
                                            tag,
                                            tag_length);
        tag_disect_tree = proto_item_add_subtree(tp, ett_wimax_tlv);
        proto_tree_add_uint(tag_disect_tree, hf_wimax_tlv_tag, tvb, offset, 2,
                            tag);
        offset = offset + 2;
        proto_tree_add_uint(tag_disect_tree, hf_wimax_tlv_length, tvb, offset,
                            2, tag_length);
        if (0 != (tag_length - 4))
        {
            offset = offset + 2;
            tag_data = g_malloc((tag_length - 4));
            tvb_memcpy(tvb, (guint8*)tag_data, offset, (tag_length - 4));
            tag_data_format = g_malloc((((tag_length - 4) * 2) + 1));
            memset(tag_data_format, '\0', (((tag_length -4) * 2) + 1));
            cursor = 0;
            while (tag_data_format_index < (tag_length - 4))
            {
                g_snprintf(&(tag_data_format[cursor]),3,"%02hhx",
                           tag_data[tag_data_format_index]);
                tag_data_format_index = tag_data_format_index + 1;
                cursor = cursor + 2;
            }
            proto_tree_add_string_format(tag_disect_tree, hf_wimax_tlv_value,
                                         tvb, offset, (tag_length - 4),
                                         tag_data, "TLV Data: %s",
                                         tag_data_format);
            g_free(tag_data);
            g_free(tag_data_format);
        }

        if (1 == get_tag_type(tag))
        {
            /* Move to children */
            if (1 == dissect_wimax_tlv(tvb, pinfo, tag_disect_tree, offset,
                                       (tag_length - 4)))
            {
                return 1;
            }
        }
        offset = offset + (tag_length - 4);
        /* Move to siblings (if any) */
        remaining_length = remaining_length - tag_length;
/*      if (0 == remaining_length)
        {
            return 0;
        }
        else
        {
            if (1 == dissect_wimax_tlv(tvb, pinfo, tree, offset,
                                        remaining_length))
            {
                return 1;
            }
        }
*/
         if (1 == dissect_wimax_tlv(tvb, pinfo, tree, offset, remaining_length))
         {
            return 1;
         }
    }
    return 0;
}

/* Code to actually dissect the packets */
static int
dissect_wimax(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{

/* Set up structures needed to add the protocol subtree and manage it */
    proto_item *wimax_item = NULL;
    proto_tree *wimax_tree = NULL;
    proto_tree *tv = NULL;
    proto_tree *tf = NULL;
    proto_tree *flag_tree = NULL;
    proto_tree *op_id_and_msg_type_tree = NULL;
    guint offset = 0;
    guint length = 0;
    guint flags = 0;
    guint op_id_and_msg_type = 0;
    guint function_type = 0;
    gchar *function_type_text = NULL;
    gchar *message_type_text = NULL;
    gchar *flags_display_str = "T and R bits";
    guint8 *msid = NULL;
    guint remaining_length = 0;

/* Make entries in Protocol column and Info column on summary display */
    if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
        col_set_str(pinfo->cinfo, COL_PROTOCOL, "WIMAX");

   /* going with the last option */
    if (check_col(pinfo->cinfo, COL_INFO)) 
        col_clear(pinfo->cinfo, COL_INFO);

    if (tree) {


/* create display subtree for the protocol */
        wimax_item = proto_tree_add_item(tree, proto_wimax, tvb, 0, -1, FALSE);
        if (wimax_item)
        {
            wimax_tree = proto_item_add_subtree(wimax_item, ett_wimax);
            if (wimax_tree)
            {
/* add an item to the subtree, see section 1.6 for more information */
                /* Version 0:1 */
                offset = 0;
                length = 1;
                proto_tree_add_item(wimax_tree, hf_wimax_version, tvb, offset,
                                    length, FALSE);
                /* Flags 1:1 */
                offset = 1;
                length = 1;
                flags = tvb_get_guint8(tvb, offset);
                tf = proto_tree_add_uint_format(wimax_tree, hf_wimax_flags, tvb,
                                                offset, length, flags, 
                                                "Flags: 0x%02x (%s)", flags,
                                                flags_display_str);
                flag_tree = proto_item_add_subtree(tf, ett_wimax_flags);
                proto_tree_add_boolean(flag_tree, hf_wimax_flags_tbit, tvb,
                                       offset, length, flags);
                proto_tree_add_boolean(flag_tree, hf_wimax_flags_rbit, tvb,
                                       offset, length, flags);
                /* Flags 1:1 */
                offset = 1;
                length = 1;
                /* Function Type 2:1 */
                offset = 2;
                length = 1;
                function_type = tvb_get_guint8(tvb, offset);
                proto_tree_add_item(wimax_tree, hf_wimax_function_type, tvb,
                                    offset, length, FALSE);
                /* OP ID and Message Type 3:1*/
                offset = 3;
                length = 1;
                op_id_and_msg_type = tvb_get_guint8(tvb, offset);
                tv = proto_tree_add_uint_format(wimax_tree,
                                           hf_wimax_op_id_and_msg_type,
                                           tvb, offset, length,
                                           op_id_and_msg_type,
                                           "Operation ID: %x, Message Type: %u",
                                           ((op_id_and_msg_type & 0xE0) >> 5),
                                           (op_id_and_msg_type & 0x1F));
                op_id_and_msg_type_tree = proto_item_add_subtree(tv,
                                                  ett_wimax_op_id_and_msg_type);
                proto_tree_add_uint(op_id_and_msg_type_tree, hf_wimax_op_id,
                                    tvb, offset, length, op_id_and_msg_type);
                proto_tree_add_uint(op_id_and_msg_type_tree, hf_wimax_msg_type,
                                    tvb, offset, length, op_id_and_msg_type);
                /* Length 4:2 */
                offset = 4;
                length = 2;
                proto_tree_add_item(wimax_tree, hf_wimax_length, tvb, offset,
                                    length, FALSE);
                /* MS ID 6:6 */
                offset = 6;
                length = 6;
                msid = g_malloc(6);
                tvb_memcpy(tvb, msid, offset, length);
                proto_tree_add_item(wimax_tree, hf_wimax_msid, tvb, offset,
                                    length, FALSE);
                /* Reserved 12:4 */
                offset = 12;
                length = 4;
                proto_tree_add_item(wimax_tree, hf_wimax_fourth_word_reserved,
                                    tvb, offset, length, FALSE);
                /* Reserved 16:2 */
                offset = 16;
                length = 2;
                proto_tree_add_item(wimax_tree, hf_wimax_transaction_id, tvb,
                                    offset, length, FALSE);
                /* Reserved 18:2 */
                offset = 18;
                length = 2;
                proto_tree_add_item(wimax_tree, hf_wimax_fifth_word_reserved,
                                    tvb, offset, length, FALSE);

                /* Lets update the Info */
                op_id_and_msg_type = (op_id_and_msg_type & 0x1F);
                if (check_col(pinfo->cinfo, COL_INFO)) {
                      get_func_type(function_type, &function_type_text);
                      get_message_type(function_type, &function_type_text,
                                       op_id_and_msg_type, &message_type_text);
                      col_add_fstr(pinfo->cinfo, COL_INFO," %s -> %s for %02hhx"
                                   ":%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
                                   function_type_text ,message_type_text ,
                                   msid[0], msid[1], msid[2], msid[3], msid[4],
                                   msid[5]); 
                }
                g_free(msid);

                /* We now disect the TLVs */

                /* Get the remaining length */
                offset = 20;
                length = 0;
                remaining_length = tvb_length_remaining(tvb, offset);
                if (4 <= remaining_length)
                {
                    /* We have data to be processed */
                    dissect_wimax_tlv(tvb, pinfo, wimax_tree, offset,
                                      remaining_length);
                }
            }
        }
    }
    else
    {
        /* Get funtion type */
        offset = 2;
        length = 1;
        function_type = tvb_get_guint8(tvb, offset);
        /* Get message type */
        offset = 3;
        length = 1;
        op_id_and_msg_type = tvb_get_guint8(tvb, offset);
        /* Get MS ID */
        offset = 6;
        length = 6;
        msid = g_malloc(6);
        tvb_memcpy(tvb, msid, offset, length);
        op_id_and_msg_type = (op_id_and_msg_type & 0x1F);
        if (check_col(pinfo->cinfo, COL_INFO)) {
              get_func_type(function_type, &function_type_text);
              get_message_type(function_type, &function_type_text,
                               op_id_and_msg_type, &message_type_text);
              col_add_fstr(pinfo->cinfo, COL_INFO," %s -> %s  for %02hhx:%02hhx"
                           ":%02hhx:%02hhx:%02hhx:%02hhx", function_type_text,
                           message_type_text, msid[0], msid[1], msid[2], 
                           msid[3], msid[4], msid[5]); 
        }
        g_free(msid);
    }

/* If this protocol has a sub-dissector call it here, see section 1.8 */

/* Return the amount of data this dissector was able to dissect */
    return tvb_length(tvb);
}


/* Register the protocol with Wireshark */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_wimax(void)
{                 
    module_t *wimax_module;

/* Setup list of header fields  See Section 1.6.1 for details*/
    static hf_register_info hf[] = {
        { &hf_wimax_version,
            { "Version",           "wimax.version",
            FT_UINT8, BASE_DEC, NULL, 0x0,          
            "Protocol Version", HFILL }
        },
        { &hf_wimax_flags,
            { "Flags",           "wimax.flags",
            FT_UINT8, BASE_HEX, NULL, 0x0,          
            "Flags", HFILL }
        },
        { &hf_wimax_flags_tbit,
            { "Flags TBIT",           "wimax.flags.tbit",
            FT_BOOLEAN, 8, TFS(&flags_set_truth), 0x01,          
            "Flags TBIT", HFILL }
        },
        { &hf_wimax_flags_rbit,
            { "Flags RBIT",           "wimax.flags.rbit",
            FT_BOOLEAN, 8, TFS(&flags_set_truth), 0x02,          
            "Flags RBIT", HFILL }
        },
        { &hf_wimax_function_type,
            { "Function Type",           "wimax.function_type",
            FT_UINT8, BASE_DEC, NULL, 0x0,          
            "Function Type", HFILL }
        },
        { &hf_wimax_op_id_and_msg_type,
            { "OP ID and Message Type",           "wimax.op_id_and_msg_type",
            FT_UINT8, BASE_HEX, NULL, 0x0,          
            "Operation ID and Message Type", HFILL }
        },
        { &hf_wimax_op_id,
            { "OP ID",           "wimax.op_id",
            FT_UINT8, BASE_HEX, NULL, 0xE0,          
            "Operation ID", HFILL }
        },
        { &hf_wimax_msg_type,
            { "Message Type",           "wimax.msg_type",
            FT_UINT8, BASE_HEX, NULL, 0x1F,          
            "Message Type", HFILL }
        },
        { &hf_wimax_length,
            { "Length",           "wimax.length",
            FT_UINT16, BASE_DEC, NULL, 0x0,          
            "Message Length", HFILL }
        },
        { &hf_wimax_msid,
            { "Mobile Station ID",           "wimax.msid",
            FT_ETHER, BASE_NONE, NULL, 0x0,          
            "Mobile Station ID", HFILL }
        },
        { &hf_wimax_fourth_word_reserved,
            { "Reserved Four Byte",           "wimax.fourth_word_reserved",
            FT_UINT32, BASE_DEC, NULL, 0x0,          
            "Reserved Foure Byte", HFILL }
        },
        { &hf_wimax_transaction_id,
            { "Transaction ID",           "wimax.transaction_id",
            FT_UINT16, BASE_DEC, NULL, 0x0,          
            "Transaction ID", HFILL }
        },
        { &hf_wimax_fifth_word_reserved,
            { "Reserved Two Byte",           "wimax.fifth_word_reserved",
            FT_UINT16, BASE_DEC, NULL, 0x0,          
            "Reserved Two Byte", HFILL }
        },
        { &hf_wimax_tlv,
            { "Wimax TLV",           "wimax.tlv",
            FT_PROTOCOL, BASE_HEX, NULL, 0x0,          
            "Wimax TLV", HFILL }
        },
        { &hf_wimax_tlv_tag,
            { "Wimax TLV tag",           "wimax.tlv.tag",
            FT_UINT16, BASE_DEC, NULL, 0x0,          
            "Wimax TLV tag", HFILL }
        },
        { &hf_wimax_tlv_length,
            { "Wimax TLV length",           "wimax.tlv.length",
            FT_UINT16, BASE_DEC, NULL, 0x0,          
            "Wimax TLV length", HFILL }
        },
        { &hf_wimax_tlv_value,
            { "Wimax TLV value",           "wimax.tlv.value",
            FT_STRING, BASE_HEX, NULL, 0x0,          
            "Wimax TLV value", HFILL }
        }
    };

/* Setup protocol subtree array */
    static gint *ett[] = {
        &ett_wimax,
        &ett_wimax_flags,       
        &ett_wimax_op_id_and_msg_type,
        &ett_wimax_tlv
    };

/* Register the protocol name and description */
    proto_wimax = proto_register_protocol("Wimax Protocol",
        "WIMAX", "wimax");

/* Required function calls to register the header fields and subtrees used */
    proto_register_field_array(proto_wimax, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));
        
/* Register preferences module (See Section 2.6 for more on preferences) */
    wimax_module = prefs_register_protocol(proto_wimax, 
        proto_reg_handoff_wimax);
     
/* Register a sample preference */
    prefs_register_uint_preference(wimax_module, "udp.wimax_port", 
         "WiMAX UDP port",
         "Set the UDP port for WiMAX messages",
         10, &global_wimax_udp_port);
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This exact format is required because a script is used to find these
   routines and create the code that calls these routines.
   
   This function is also called by preferences whenever "Apply" is pressed 
   (see prefs_register_protocol above) so it should accommodate being called 
   more than once.
*/
void
proto_reg_handoff_wimax(void)
{
    static gboolean inited = FALSE;
    static int currentPort = -1;
        
    if (!inited) {


/*  Use new_create_dissector_handle() to indicate that dissect_wimax()
 *  returns the number of bytes it dissected (or 0 if it thinks the packet
 *  does not belong to wimax).
 */
        wimax_handle = new_create_dissector_handle(dissect_wimax,
            proto_wimax);
        inited = TRUE;
    }
        
        /* 
          If you perform registration functions which are dependant upon
          prefs the you should de-register everything which was associated
          with the previous settings and re-register using the new prefs
      settings here. In general this means you need to keep track of what
      value the preference had at the time you registered using a local
      static in this function. ie.

        */

          if (currentPort != -1) {
              dissector_delete("udp.port", currentPort, wimax_handle);
          }

          currentPort = global_wimax_udp_port;
          dissector_add("udp.port", currentPort, wimax_handle);
}
